@@ -6,7 +6,7 @@ from django.db.transaction import atomic, Atomic
 from django.db.utils import OperationalError, IntegrityError
 
 from ..backend import Backend, Lock, NotFound, NoResult, Failed
-from ..contrib.django.models import TaskkitControlEvent, TaskkitLock, TaskkitWorker, TaskkitTask, TaskkitSchedulerState
+from ..contrib.django.models import TaskkitControlEvent, TaskkitLock, TaskkitWorker, TaskkitTask, TaskkitTaskQueue, TaskkitSchedulerState
 from ..event import EventBridge, ControlEvent, encode_control_event, decode_control_event
 from ..kit import Kit
 from ..stage import StageInfo
@@ -48,7 +48,7 @@ class DjangoLock(Lock):
             try:
                 with atomic():
                     TaskkitLock.objects\
-                        .select_for_update(nowait=True)\
+                        .select_for_update(skip_locked=True)\
                         .get(pk=self.target)
             except TaskkitLock.DoesNotExist:
                 try:
@@ -62,7 +62,7 @@ class DjangoLock(Lock):
         self.atomic.__enter__()
         try:
             TaskkitLock.objects\
-                .select_for_update(nowait=True)\
+                .select_for_update(skip_locked=True)\
                 .get(pk=self.target)
             self.acquired = True
             return True
@@ -105,8 +105,17 @@ class DjangoBackend(Backend):
             return
         objects = [self._task_to_db(t) for t in tasks]
         with atomic():
-            self.discard_tasks(*[t.pk for t in objects])
             TaskkitTask.objects.bulk_create(objects, ignore_conflicts=True)
+            TaskkitTaskQueue.objects.bulk_create([
+                self._db_task_to_queue(t) for t in objects
+            ], ignore_conflicts=True)
+
+    def retry_task(self, task: Task):
+        with atomic():
+            self.discard_tasks(task.id)
+            db_task = self._task_to_db(task)
+            db_task.save()
+            self._db_task_to_queue(db_task).save()
 
     def get_queued_tasks(self, group: str, limit: int) -> list[Task]:
         return [
@@ -123,26 +132,39 @@ class DjangoBackend(Backend):
         return {tid: tasks.get(tid) for tid in task_ids}
 
     def assign_task(self, group: str, worker_id: str) -> Optional[Task]:
-        for pk in TaskkitTask.objects\
-                .filter(began__isnull=True, group=group, due__lt=cur_ts())\
-                .values_list('pk', flat=True)\
-                .order_by('due')[:50]:
-            with atomic():
-                try:
-                    db_task = TaskkitTask.objects\
-                        .select_for_update(nowait=True)\
-                        .get(pk=pk)
-                    if db_task.began is None:
-                        db_task.assignee_worker_id = worker_id
-                        db_task.began = cur_ts()
-                        db_task.save()
-                        return self._db_to_task(db_task)
-                except (OperationalError, TaskkitTask.DoesNotExist):
-                    pass
+        n = 8
+        while True:
+            pks = list(
+                TaskkitTaskQueue.objects
+                .filter(group=group, due__lt=cur_ts())
+                .values_list('pk', flat=True)
+                .order_by('due')[:n])
+            for pk in pks:
+                if (task := self._assign(pk, worker_id)):
+                    return task
+            if len(pks) < n:
+                return None
+            n *= 2
+
+    def _assign(self, task_id: str, worker_id: str) -> Optional[Task]:
+        with atomic():
+            try:
+                db_task = TaskkitTask.objects\
+                    .select_for_update(skip_locked=True)\
+                    .get(pk=task_id)
+                TaskkitTaskQueue.objects.filter(id=task_id).delete()
+                if db_task.began is None:
+                    db_task.assignee_worker_id = worker_id
+                    db_task.began = cur_ts()
+                    db_task.save()
+                    return self._db_to_task(db_task)
+            except (OperationalError, TaskkitTask.DoesNotExist):
+                pass
         return None
 
     def discard_tasks(self, *task_ids: str):
         TaskkitTask.objects.filter(pk__in=task_ids).delete()
+        TaskkitTaskQueue.objects.filter(pk__in=task_ids).delete()
 
     def succeed(self, task: Task, result: bytes):
         with atomic():
@@ -222,8 +244,8 @@ class DjangoBackend(Backend):
                 .filter(began__isnull=False)\
                 .get(pk=info.task_id)
             task.began = None
-            task.assignee_worker_id = None
             task.save()
+            self._db_task_to_queue(task).save()
         except TaskkitTask.DoesNotExist:
             pass
 
@@ -286,6 +308,9 @@ class DjangoBackend(Backend):
             disposable=None,
         )
 
+    def _db_task_to_queue(self, t: TaskkitTask) -> TaskkitTaskQueue:
+        return TaskkitTaskQueue(id=t.id, group=t.group, due=t.due)
+
 
 def make_kit(handler: TaskHandler) -> Kit:
     return Kit(DjangoBackend(), DjangoEventBridge(), handler)
